<input type="text" id="colors-0" name="colors-0" value="${data['colors'][0] if data else 'A color'}" />
<input type="text" id="colors-1" name="colors-1" value="${data['colors'][1] if data else 'A color'}" />